#include <math.h>
#include <float.h>
#include <stdio.h>

#include "grind.h"


double length(double, double*, double, double*);
double min(double, double);

boolean trickOutside(int mode,double a,double b,double c) {
  /* check whether the arguments lie outside the state-space */ 

  if (a < ax2val(0.0, X_AXIS)-TINY || a > ax2val(1.0, X_AXIS)+TINY ||
      b < ax2val(0.0, Y_AXIS)-TINY || b > ax2val(1.0, Y_AXIS)+TINY ||
      ((DIM3 == mode) &&
       (c < ax2val(0.0, Z_AXIS)-TINY || c > ax2val(1.0, Z_AXIS)+TINY)))
    return true;                                                                                
  return false;
}

int contEqui(int cont_ax, double arc_length, int direction) {
  static double *state0 = NULL;/*  state vector of previous state */
  double *state1; /*  state vector of current state; points to fstate */
  static double *dstate0 = NULL, *dstate1= NULL; /* first derivative vectors */
  double p0, p1; /* parameter value of last and previous step */
  double dp0, dp1; /* first derivs of parameter value of last and prev step */
  int stab0 = 0, stab1 = 0; /* stability of last and previous step */

  double len; /*  realized arclength */
  double safety = 2.;  /*  safety in calculating projection next state */
  int par; /*  the parameter we're continuing with */
  boolean error = false;   /*  error flag */
  double init_step = 1.;   /*  init stepsize for finding first two points */
  int i, count;/*  counters */
  const int max_count = 10; /*  max count for no of projection trials */
  const int newtrls = 25;  /*  no. of newton trials */
  extern double tolnew; /*  used in mnewt (defined in commands.c) */
  int mode; /* for 2D continuations in 3D space */

  if (Interrupt()) return 0;
  if (DIM2 == mode_opt && Z_AXIS == cont_ax) 
    mode = DIM3;
  else
    mode = mode_opt;
 
  /*  first time memory allocation */
  if (NULL == state0) {
    state0 = dcalloc(nvar, "'state0' in continue()");
    dstate0 = dcalloc(nvar, "'dstate0' in continue()");
    dstate1 = dcalloc(nvar, "'dstate1' in continue()");
  }
  state1 = fstate; /* state1 holds the state of the current step, ie. fstate */
  /*  which parameter are we continuing with? */
  par = axis[cont_ax].var;
  /*  and check */
  if (par < nvar + nexpr)
    return gr_warning("not a parameter: %s\n", 
	       (par>=0 && par<nnames)?all_names[par]:"?");
  /* initialize first step size; misusing p0, p1 */
  p0 = all_values[par];
  p1 = ax2val(1.01*val2ax(p0, cont_ax), cont_ax);
  init_step = min(fabs(p1-p0), arc_length/safety);
  /* initialize first point (remember original par value; restore at end) */
  storeaxis(true);
  p0 = p1 = all_values[par];
  dp0 = dp1 = 0.0;
  for (i=0; i<nvar; i++)
    state0[i] = state1[i] = nstate[i];

  /*  newton to make sure we are in an equilibrium */
  /*  mnewt() is NumRec so we give "state0-1" as argument */
  if (0 == mnewt(newtrls, state1 - 1, nvar, tolnew, tolnew)) {
    error = true;
    gr_warning("No initial convergence!\n");
  }
  else {
    if (trickOutside(mode,all_values[axis[0].var],all_values[axis[1].var],
		all_values[axis[2].var]))
      gr_warning("Initial point not in diagram\n");
    /*  determine stability of equilibrium */
    stab1 = stab0 = eigen(state1, false);
    /*  start drawing the trajectory */
    start_continuation((STABLE == stab1));
    count = 10;
    while(count--) {
      all_values[par] = p1 = p0 + init_step*direction;
      if (0 == mnewt(newtrls, state1-1, nvar, tolnew, tolnew))
	init_step *= .1;
      else {
	/*  determine first arc_length */
	len = length(p0, state0, p1, state1);
	/*  is it within limits? */
	if(len <= arc_length) {
	  /*  set first derivatives for two steps */
	  /*  now the first step in continuation below will be a linear step */
	  dp0 = dp1 = (p1 - p0)*arc_length/len;
	  for (i=0; i<nvar; i++)
	    dstate0[i] = dstate1[i] = (state1[i]-state0[i])*arc_length/len;
	  stab1 = eigen(state1, false);
	  add2trajectory();
	  if (stab0 != stab1) { /*  stability changed! */
	    if (mark_opt)
	      setMark(fstate, 0);
	    /* start new curve with appropriate color */
	    start_continuation((STABLE == stab1));
	    printf("Bifurcation at %s = ", all_names[par]);
	    printf(dformat, p1);
	    printf("\n");
	  }
	  break; /* first point accepted; go on */
	}
	else
	  init_step *= 0.1;
      }
    }
    if (0 == count)
      error = true;
  }

  /*  start the actual continuation */
  while (!error && !trickOutside(mode,all_values[axis[0].var],
				 all_values[axis[1].var],
				 all_values[axis[2].var])) {
    /*  find next point and determine new dp1 and new dstate1 */
    /*  first copy last state into previous state and */
    /*  estimate new point using linear projection and the safety factor; */
    p0 = p1;
    for (i=0; i<nvar; i++)
      state0[i] = state1[i];
    stab0 = stab1;
    count = max_count;
    while(count--) {
      if (Interrupt())
	goto exit;
      all_values[par] = p1 = p0 + dp1/safety + (dp1 - dp0)/(safety*2.);
      if (fabs(p1 - p0) < TINY) {
	error = true;
	gr_warning("too small step size in continue\n");
	break;
      }
      for (i=0; i<nvar; i++)
	state1[i] = state0[i] + 
	  dstate1[i]/safety + (dstate1[i] - dstate0[i])/(safety*2.);
      if (mnewt(newtrls, state1-1, nvar, tolnew, tolnew)) {
	/*  convergence; check new arc-length */
	len = length(p0, state0, p1, state1);
	if (len <= arc_length) {
	  /*  complete succes; reset safety */
	  safety = 2.;
	  dp0 = dp1;
	  dp1 = (p1 - p0)*arc_length/len;
	  for (i=0; i<nvar; i++) {
	    dstate0[i] = dstate1[i];
	    dstate1[i] = (state1[i]-state0[i])*arc_length/len;
	  }
	  stab1 = eigen(state1, false);	  
	  add2trajectory();
	  if (stab0 != stab1) { /*  stability changed! */
	    if (mark_opt)
	      setMark(fstate, 0);
	    start_continuation((STABLE == stab1));
	    printf("Bifurcation at %s = ", all_names[par]);
	    printf(dformat, p1);
	    printf("\n");
	  }
	  break; /* point accepted; go on */
	}
      }
      /*  no convergence or too big a step */
      /*  adjust safety parameter (ie make a smaller step) and redo */
      safety *= 2.;
    }
    if (!count) {
      gr_warning("no convergence in continue\n");
      error = true;
      break;
    }
  }
  if (error) {
    printf("No convergence: \n");
    printf("Last point at: \n");
    printf(sformat, pars[par-nvar-nexpr].name);
    printf(" = ");
    printf(dformat, all_values[par]);
    printf("\n");
    prider (state1);
  }
  /*  reset original par and var values */
 exit:
  storeaxis(false);
  return !Interrupt();
}

double length(double p0, double *state0, double p1, double *state1) {
  double s;
  int i;

  s = (p1-p0)*(p1-p0);
  for (i=0; i<nvar; i++)
    s += (state1[i]-state0[i]) * (state1[i]-state0[i]);
  return sqrt(s);
}
